#pragma once

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <infiniband/verbs.h>

#include "libcapulet/net_udp.h"

struct CapuletRdmaIbContext {
	struct ibv_context *ib_ctx;
	struct ibv_pd *pd;
	struct ibv_cq *send_cq;
	struct ibv_cq *recv_cq;
	struct ibv_qp *qp;
};

bool capulet_rdma_ib_initialize_device(struct CapuletRdmaIbContext *ctx, const char *device_name);
bool capulet_rdma_ib_initialize_qp(struct CapuletRdmaIbContext *ctx, int access);
struct ibv_mr *capulet_rdma_ib_create_mr(struct CapuletRdmaIbContext *ib_ctx, size_t size, int flags);
bool capulet_rdma_ib_fill_base_udp(struct CapuletRdmaIbContext *ctx, struct CapuletNetUdpQpInfoPacket *own);
bool capulet_rdma_ib_set_peer_from_udp(struct CapuletRdmaIbContext *ctx, struct CapuletNetUdpContext *udp_ctx);
bool capulet_rdma_ib_post_recv(struct CapuletRdmaIbContext *ctx, struct ibv_mr *ibv_dev_mr, int mr_size, uint32_t id_prefix);
bool capulet_rdma_ib_post_send(struct CapuletRdmaIbContext *ctx, enum ibv_wr_opcode opcode, struct ibv_mr *local_mr, int local_mr_size, uint32_t id_prefix, struct CapuletNetUdpMrInfoPacket *remote_mr);
void capulet_rdma_ib_free_mr(struct ibv_mr *mr);
void capulet_rdma_ib_free(struct CapuletRdmaIbContext *ctx);

