#include "common/hashtable.h"

#include <stdio.h>
#include <stdlib.h>

#define POTOML_MAX_PROPERTY_LENGTH	COMMON_HASHTABLE_KEY_LENGTH

enum PotomlError {
	InvalidCharacter = 10,
	MissingClosingBracket = 20,
	KeyWithoutValue = 30,
	HashtableInsertError = 40,
	FileDoesNotExist = 50,
	CouldNotReadFromFile = 51,
};

enum PotomlState {
	PotomlLineBeginning = 0x00,
	PotomlLineEnding = 0x01,
	PotomlCommentContent = 0x10,
	PotomlSectionName = 0x20,
	PotomlPropertyName = 0x30,
	PotomlAfterPropertyName = 0x31,
	PotomlBeforePropertyValue = 0x32,
	PotomlPropertyValueString = 0x40,
	PotomlPropertyValueLong = 0x41,
	PotomlPropertyValueDouble = 0x42,
};

struct PotomlTomlSubset {
	struct CommonHashtableTable *hashtable;
	char *section_name;
	size_t section_name_length;
};

char potoml_toml_parse(struct CommonHashtableTable *hashtable, char *toml);
char potoml_toml_parse_file(struct CommonHashtableTable *hashtable, char *toml_file_path);
struct PotomlTomlSubset *potoml_toml_create_subset(struct CommonHashtableTable *hashtable, const char *section_name);
char *potoml_toml_get_string(struct PotomlTomlSubset *subset, const char *property);
double potoml_toml_get_double(struct PotomlTomlSubset *subset, const char *property);
long potoml_toml_get_long(struct PotomlTomlSubset *subset, const char *property);
void potoml_toml_subset_free(struct PotomlTomlSubset *subset);
void potoml_toml_free(struct CommonHashtableTable *hashtable);

