#pragma once

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <pthread.h>

#include "amarisoft/trx_driver.h"
#include "common/circular_buffer.h"
#include "libcapulet/rdma_ib.h"
#include "libcapulet/net_udp.h"

#define TRX_RDMA_MR_COUNT 32
#define TRX_RDMA_MR_SIZE 256
#define TRX_RDMA_BUFFER_SIZE 131072
// XXX: hardcoded parameter
#define TRX_RDMA_CHANNEL_COUNT 2

struct SDRMetadata {
	uint8_t version;
	uint16_t sample_count;
	uint32_t frame_number;
	uint8_t __unused[25];
};

struct SDRMemoryRegion {
	struct SDRMetadata meta;
	TRXComplex iq[TRX_RDMA_CHANNEL_COUNT][TRX_RDMA_MR_SIZE];
};

struct SDRContext {
	struct CapuletRdmaIbContext ib_ctx;
	struct CapuletNetUdpContext *udp_ctx;

	struct CommonCBBuffer *recv_buffer;

	trx_timestamp_t last_timestamp;

	struct ibv_mr *recv_mrs[TRX_RDMA_MR_COUNT];
	struct ibv_mr *send_mrs[TRX_RDMA_MR_COUNT];

	uint64_t send_mrs_status;
	uint32_t frame_count;
	int consumed_mr_id;
	int consumed_mr_length;

	char *server_addr;
	char *device_name;
};

