#include "trx-bridge/bridge.h"

char *bridge_trx_input_get_param_string(void *app_opaque, const char *prop_name) {
	return potoml_toml_get_string((struct PotomlTomlSubset *) app_opaque, prop_name);
}

int bridge_trx_input_get_param_double(void *app_opaque, double *pval, const char *prop_name) {
	return potoml_toml_get_double((struct PotomlTomlSubset *) app_opaque, prop_name);
}

char *bridge_trx_output_get_param_string(void *app_opaque, const char *prop_name) {
	return potoml_toml_get_string((struct PotomlTomlSubset *) app_opaque, prop_name);
}

int bridge_trx_output_get_param_double(void *app_opaque, double *pval, const char *prop_name) {
	return potoml_toml_get_double((struct PotomlTomlSubset *) app_opaque, prop_name);
}

int bridge_trx_pthread_init(void *app_opaque2, const TRXPThreadConfig *config) {
	// XXX: this function should likely join with the main thread
	return 0;
}

void bridge_trx_pthread_terminate(void *app_opaque2) {
}

TRXDriverParams2 *fill_trx_driver_params(struct PotomlTomlSubset *config) {
	TRXDriverParams2 *trx_driver_params = malloc(sizeof(TRXDriverParams2));

	int rxi, txi, pi, ci;
	char param_name[256];

	trx_driver_params->rx_channel_count = potoml_toml_get_long(config, "rx_channel_count");
	trx_driver_params->tx_channel_count = potoml_toml_get_long(config, "tx_channel_count");
	trx_driver_params->rf_port_count = potoml_toml_get_long(config, "rf_port_count");
	trx_driver_params->cell_count = potoml_toml_get_long(config, "cell_count");
	trx_driver_params->s72_enable = potoml_toml_get_long(config, "s72_enable");

	trx_driver_params->rx_freq = malloc(trx_driver_params->rx_channel_count * sizeof(int64_t));
	trx_driver_params->rx_gain = malloc(trx_driver_params->rx_channel_count * sizeof(double));
	trx_driver_params->rx_bandwidth = malloc(trx_driver_params->rx_channel_count * sizeof(int));
	trx_driver_params->rx_port_channel_count = malloc(trx_driver_params->rx_channel_count * sizeof(int));

	for(rxi = 0; rxi < trx_driver_params->rx_channel_count; rxi++) {
		sprintf(param_name, "rx%d.freq", rxi);
		trx_driver_params->rx_freq[rxi] = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "rx%d.gain", rxi);
		trx_driver_params->rx_gain[rxi] = potoml_toml_get_double(config, param_name);

		sprintf(param_name, "rx%d.bandwidth", rxi);
		trx_driver_params->rx_bandwidth[rxi] = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "rx%d.port_channel_count", rxi);
		trx_driver_params->rx_port_channel_count[rxi] = potoml_toml_get_long(config, param_name);
	}

	trx_driver_params->tx_freq = malloc(trx_driver_params->tx_channel_count * sizeof(int64_t));
	trx_driver_params->tx_gain = malloc(trx_driver_params->tx_channel_count * sizeof(double));
	trx_driver_params->tx_bandwidth = malloc(trx_driver_params->tx_channel_count * sizeof(int));
	trx_driver_params->tx_port_channel_count = malloc(trx_driver_params->tx_channel_count * sizeof(int));

	for(txi = 0; txi < trx_driver_params->tx_channel_count; txi++) {
		sprintf(param_name, "tx%d.freq", txi);
		trx_driver_params->tx_freq[txi] = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "tx%d.gain", txi);
		trx_driver_params->tx_gain[txi] = potoml_toml_get_double(config, param_name);

		sprintf(param_name, "tx%d.bandwidth", txi);
		trx_driver_params->tx_bandwidth[txi] = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "tx%d.port_channel_count", txi);
		trx_driver_params->tx_port_channel_count[txi] = potoml_toml_get_long(config, param_name);
	}

	trx_driver_params->sample_rate = malloc(trx_driver_params->rf_port_count * sizeof(TRXFraction));

	for(pi = 0; pi < trx_driver_params->rf_port_count; pi++) {
		sprintf(param_name, "port%d.sample_rate_num", pi);
		trx_driver_params->sample_rate[pi].num = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "port%d.sample_rate_den", pi);
		trx_driver_params->sample_rate[pi].den = potoml_toml_get_long(config, param_name);
	}

	trx_driver_params->cell_info = malloc(trx_driver_params->cell_count * sizeof(TRXCellInfo));

	for(ci = 0; ci < trx_driver_params->cell_count; ci++) {
		sprintf(param_name, "cell%d.rf_port_index", ci);
		trx_driver_params->cell_info[ci].rf_port_index = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.dl_earfcn", ci);
		trx_driver_params->cell_info[ci].dl_earfcn = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.ul_earfcn", ci);
		trx_driver_params->cell_info[ci].ul_earfcn = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.n_rb_dl", ci);
		trx_driver_params->cell_info[ci].n_rb_dl = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.n_rb_ul", ci);
		trx_driver_params->cell_info[ci].n_rb_ul = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.dl_cyclic_prefix", ci);
		trx_driver_params->cell_info[ci].dl_cyclic_prefix = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.ul_cyclic_prefix", ci);
		trx_driver_params->cell_info[ci].ul_cyclic_prefix = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.type", ci);
		trx_driver_params->cell_info[ci].type = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.tdd.uldl_config", ci);
		trx_driver_params->cell_info[ci].u.tdd.uldl_config = potoml_toml_get_long(config, param_name);

		sprintf(param_name, "cell%d.tdd.special_subframe_config", ci);
		trx_driver_params->cell_info[ci].u.tdd.special_subframe_config = potoml_toml_get_long(config, param_name);
	}

	return trx_driver_params;
}

int main(int argc, char *argv[]) {
	struct CommonHashtableTable config;
	struct PotomlTomlSubset *bridge_config;
	struct PotomlTomlSubset *input_driver_config;
	struct PotomlTomlSubset *output_driver_config;
	struct PotomlTomlSubset *output_driver_start_config;

	TRXState input_trx_state;
	TRXState output_trx_state;
	TRXDriverParams2 *output_trx_start_params;
	TRXReadMetadata read_md;
	TRXWriteMetadata write_md;

	trx_timestamp_t input_rx_ts = 0;
	trx_timestamp_t input_tx_ts = 0;
	trx_timestamp_t output_rx_ts = 0;
	trx_timestamp_t output_tx_ts = 0;

	size_t bridge_config_path_size;
	size_t driver_config_dir_size;

	int opt;
	int input_read_count;
	int output_read_count;
	int input_max_read_count = 0;
	int result;
	char *error;

	char *output_lib_name = NULL;
	char *input_lib_name = NULL;
	char *bridge_config_path = NULL;
	char *driver_config_dir = NULL;

	void *input_lib_handle;
	void *output_lib_handle;

	int (*input_lib_init)(TRXState *s);
	int (*output_lib_init)(TRXState *s);

	// IQ samples are complex numbers
	// A flat plane of Complex is created for all channels
	TRXComplex *input_samples = malloc(2 * TRX_BUFFER_SIZE * sizeof(TRXComplex));
	TRXComplex *input_channels[2] = { input_samples, &input_samples[TRX_BUFFER_SIZE] };

	TRXComplex *output_samples = malloc(2 * TRX_BUFFER_SIZE * sizeof(TRXComplex));
	TRXComplex *output_channels[2] = { output_samples, &output_samples[TRX_BUFFER_SIZE] };

	while((opt = getopt(argc, argv, "c:p:")) != -1) {
		if(opt == 'c') {
			bridge_config_path_size = strlen(optarg) + 1;
			bridge_config_path = (char *) malloc(bridge_config_path_size * sizeof(char));
			strcpy(bridge_config_path, optarg);
		} else if(opt == 'p') {
			driver_config_dir_size = strlen(optarg) + 1;
			driver_config_dir = (char *) malloc(driver_config_dir_size * sizeof(char));
			strcpy(driver_config_dir, optarg);
		}
	}

	if(!bridge_config_path || !driver_config_dir) {
		fprintf(stderr, "Missing required options: -c config_file and -p driver_config_dir\n");
		return -1;
	}

	// Before starting anything, the process should be made as high
	// priority as possible
	if(nice(-20) != -20) {
		fprintf(stderr, "Setting priority failed, run this program as root\n");
		return -1;
	}

	printf("trx-bridge: reading configuration file %s\n", bridge_config_path);

	common_hashtable_initialize(&config, 255);
	result = potoml_toml_parse_file(&config, bridge_config_path);

	if(result != 0) {
		fprintf(stderr, "Reading configuration file returned 0x%x\n", result);
		return -1;
	}

	bridge_config = potoml_toml_create_subset(&config, "bridge");
	input_driver_config = potoml_toml_create_subset(&config, "input-driver");
	output_driver_config = potoml_toml_create_subset(&config, "output-driver");
	output_driver_start_config = potoml_toml_create_subset(&config, "output-driver-extra");

	input_lib_name = potoml_toml_get_string(bridge_config, "input_driver_path");
	output_lib_name = potoml_toml_get_string(bridge_config, "output_driver_path");

	if(!input_lib_name || !output_lib_name) {
		fprintf(stderr, "Could not find configuration parameters for input and output drivers path\n");
		return -1;
	}

	printf("trx-bridge: opening input library %s\n", input_lib_name);

	input_lib_handle = dlopen(input_lib_name, RTLD_NOW);
	if(!input_lib_handle) {
		fputs(dlerror(), stderr);
		fputs("\n", stderr);
		return -1;
	}

	// NULL is a possible return value, so test errors using `dlerror`
	input_lib_init = dlsym(input_lib_handle, "trx_driver_init");
	if((error = dlerror()) != NULL)  {
		fputs(error, stderr);
		return -1;
	}

	printf("trx-bridge: opening output library %s\n", output_lib_name);

	output_lib_handle = dlopen(output_lib_name, RTLD_NOW);
	if(!output_lib_handle) {
		fputs(dlerror(), stderr);
		fputs("\n", stderr);
		return -1;
	}

	output_lib_init = dlsym(output_lib_handle, "trx_driver_init");
	if((error = dlerror()) != NULL)  {
		fputs(error, stderr);
		return -1;
	}

	// Set required parameters, and nothing more
	printf("trx-bridge: setting driver configuration directory to %s\n", driver_config_dir);
	input_trx_state.trx_api_version = 15;
	input_trx_state.app_opaque = input_driver_config;
	input_trx_state.trx_get_param_string = &bridge_trx_input_get_param_string;
	input_trx_state.trx_get_param_double = &bridge_trx_input_get_param_double;
	input_trx_state.trx_pthread_init = &bridge_trx_pthread_init;
	input_trx_state.trx_pthread_terminate = &bridge_trx_pthread_terminate;

	output_trx_state.trx_api_version = 15;
	output_trx_state.app_opaque = output_driver_config;
	output_trx_state.trx_get_param_string = &bridge_trx_output_get_param_string;
	output_trx_state.trx_get_param_double = &bridge_trx_output_get_param_double;
	output_trx_state.trx_pthread_init = &bridge_trx_pthread_init;
	output_trx_state.trx_pthread_terminate = &bridge_trx_pthread_terminate;

	// XXX: Amarisoft stupid or purposedly flawed parameters
	input_trx_state.path = driver_config_dir;
	input_trx_state.reserved[121] = (void *) 0x24535258;
	input_trx_state.reserved[122] = (void *) 0x12;

	output_trx_state.path = driver_config_dir;
	output_trx_state.reserved[121] = (void *) 0x24535258;
	output_trx_state.reserved[122] = (void *) 0x12;

	printf("trx-bridge: initializing drivers\n");
	result = (*input_lib_init)(&input_trx_state);
	if(result != 0) {
		fprintf(stderr, "Initializing input driver failed, code %d\n", result);
		return -1;
	}

	result = (*output_lib_init)(&output_trx_state);
	if(result != 0) {
		fprintf(stderr, "Initializing output driver failed, code %d\n", result);
		return -1;
	}

	printf("trx-bridge: starting drivers\n");
	result = input_trx_state.trx_start_func2(&input_trx_state, NULL);
	if(result != 0) {
		fprintf(stderr, "Starting input driver failed, code %d\n", result);
		return -1;
	}

	// XXX: genericity should also fill this struct for input
	output_trx_start_params = fill_trx_driver_params(output_driver_start_config);
	result = output_trx_state.trx_start_func2(&output_trx_state, output_trx_start_params);
	if(result != 0) {
		fprintf(stderr, "Starting output driver failed, code %d\n", result);
		return -1;
	}

	printf("trx-bridge: started successfully\n");

	while(1) {
		// Reading from output indicates the maximum sample count to be read from input
		output_read_count = output_trx_state.trx_read_func2(&output_trx_state, &output_rx_ts, (void **) output_channels, TRX_BUFFER_SIZE, 0, &read_md);
		input_max_read_count += output_read_count;

		// Expected behavior: TX (i.e. write) timestamp is set once based on RX (i.e. read)
		// timestamp and incremented after every write.
		// The buffer size in TRX SDR is four times the samples rate, so this
		// value is used here as an offset between the two timestamps.
		if(output_tx_ts == 0) output_tx_ts = output_rx_ts + 4 * SAMPLE_RATE_KHZ;
		output_tx_ts += output_read_count;

		// Real behavior: sometimes, the TX timestamp mismatches, so it should be matched again.
		if(output_tx_ts != output_rx_ts + 4 * SAMPLE_RATE_KHZ) {
			printf("trx-bridge: output timestamp mismatch %lu != %lu\n", output_tx_ts, output_rx_ts + 4 * SAMPLE_RATE_KHZ);
			output_tx_ts = output_rx_ts + 4 * SAMPLE_RATE_KHZ;
		}

		// Write back samples read from output
		if(output_read_count > 0) input_trx_state.trx_write_func2(&input_trx_state, input_tx_ts, (const void **) output_channels, output_read_count, 0, &write_md);

		// Reading from input decrease the sample count to be read on next iteration
		input_read_count = MIN(TRX_BUFFER_SIZE, input_max_read_count);
		input_read_count = input_trx_state.trx_read_func2(&input_trx_state, &input_rx_ts, (void **) input_channels, input_read_count, 0, &read_md);
		input_max_read_count -= input_read_count;

		// Same timestamp handling as for output
		if(input_tx_ts == 0) input_tx_ts = input_rx_ts + 4 * SAMPLE_RATE_KHZ;
		input_tx_ts += input_read_count;

		if(input_tx_ts != input_rx_ts + 4 * SAMPLE_RATE_KHZ) {
			printf("trx-bridge: input timestamp mismatch %lu != %lu\n", input_tx_ts, input_rx_ts + 4 * SAMPLE_RATE_KHZ);
			input_tx_ts = input_rx_ts + 4 * SAMPLE_RATE_KHZ;
		}

		// Finally write back to output samples from input
		if(input_read_count > 0) output_trx_state.trx_write_func2(&output_trx_state, output_tx_ts, (const void **) input_channels, input_read_count, 0, &write_md);
	}

	dlclose(input_lib_handle);
	dlclose(output_lib_handle);

	potoml_toml_subset_free(bridge_config);
	potoml_toml_subset_free(input_driver_config);
	potoml_toml_subset_free(output_driver_config);
	// XXX: Amarisoft's trx_sdr frees string parameters being fed to it...
	//potoml_toml_free(&config);
	free(input_lib_name);
	free(output_lib_name);
	common_hashtable_free(&config);

	free((void *) bridge_config_path);
	free((void *) driver_config_dir);

	return 0;
}

